@@ -2,6 +2,9 @@ package JSON::Validator::OpenAPI::Mojolicious;
 use Mojo::Base 'JSON::Validator';
 
 use Carp 'confess';
+use JSON::Validator::Util qw(schema_type);
+use Mojo::JSON qw(false true);
+use Mojo::Parameters;
 use Mojo::Util;
 use Scalar::Util 'looks_like_number';
 use Time::Local ();
@@ -9,8 +12,7 @@ use Time::Local ();
 use constant DEBUG   => $ENV{JSON_VALIDATOR_DEBUG} || 0;
 use constant IV_SIZE => eval 'require Config;$Config::Config{ivsize}';
 
-our %COLLECTION_RE = (pipes => qr{\|}, csv => qr{,}, ssv => qr{\s}, tsv => qr{\t});
-our %VERSIONS      = (
+our %VERSIONS = (
   v2 => 'http://swagger.io/v2/schema.json',
   v3 => 'https://spec.openapis.org/oas/3.0/schema/2019-04-02'
 );
@@ -19,30 +21,6 @@ has version => 2;
 
 sub E { JSON::Validator::Error->new(@_) }
 
-has generate_definitions_path => sub {
-  my $self = shift;
-  Scalar::Util::weaken($self);
-
-  return sub {
-    my $ref = shift;
-
-    if ($self->version eq '3') {
-
-      # Try to determine the path from the fqn
-      # We are only interested in the path in the fqn, so following fqn:
-      #
-      # #/components/schemas/some_schema, the returned path with be ['components', 'schemas']
-      my $path = Mojo::Path->new($ref->fqn =~ m!^.*#/(components/.+)$!)->to_dir->parts;
-      return $path->[0] ? $path : ['definitions'];
-    }
-    else {
-
-      # By default return definitions as path
-      return ['definitions'];
-    }
-  };
-};
-
 sub load_and_validate_schema {
   my ($self, $spec, $args) = @_;
 
@@ -55,7 +33,7 @@ sub load_and_validate_schema {
   my @errors;
   my $gather = sub {
     push @errors, E($_[1], 'Only one parameter can have "in":"body"')
-      if 1 < grep { $_->{in} eq 'body' } @{$_[0] || []};
+      if 1 < grep { $_->{in} and $_->{in} eq 'body' } @{$_[0] || []};
   };
 
   $self->_get($self->_resolve($spec), ['paths', undef, undef, 'parameters'], '', $gather);
@@ -91,15 +69,15 @@ sub validate_request {
 
   for my $p (@{$schema->{parameters} || []}) {
     my ($in, $name, $type) = @$p{qw(in name type)};
-    $type ||= $p->{schema}{type} if $p->{schema}; # v3
     my ($exists, $value) = (0, undef);
+    $type ||= schema_type($p->{schema} || $p);
 
     if ($in eq 'body') {
-      $value = $self->_get_request_data($c, $in);
+      $value  = $self->_get_request_data($c, $in);
       $exists = length $value if defined $value;
     }
     elsif ($in eq 'formData' and $type eq 'file') {
-      $value = $self->_get_request_uploads($c, $name)->[-1];
+      $value  = $self->_get_request_uploads($c, $name)->[-1];
       $exists = $value ? 1 : 0;
     }
     else {
@@ -109,32 +87,23 @@ sub validate_request {
       $value  = $value->{$key};
     }
 
-    if (defined $value and $type eq 'array') {
+    if ($in ne 'body' and $type eq 'array') {
       $value = $self->_coerce_by_collection_format($value, $p);
     }
 
-    ($exists, $value) = (1, $p->{schema}{default})
-      if !$exists and $p->{schema} and exists $p->{schema}{default};
-    ($exists, $value) = (1, $p->{default})
-      if !$exists and exists $p->{default};
-
-    if ($type and defined $value) {
-      if ($type ne 'array' and ref $value eq 'ARRAY') {
-        $value = $value->[-1];
-      }
-      if (($type eq 'integer' or $type eq 'number') and Scalar::Util::looks_like_number($value)) {
-        $value += 0;
-      }
-      elsif ($type eq 'boolean') {
-        if (!$value or $value =~ /^(?:false)$/) {
-          $value = Mojo::JSON->false;
-        }
-        elsif ($value =~ /^(?:1|true)$/) {
-          $value = Mojo::JSON->true;
-        }
-      }
+    if ($in ne 'body' and $type eq 'object') {
+      $value  = $self->_coerce_object_by_style($c, $value, $p);
+      $exists = defined $value ? 1 : 0;
     }
 
+    ($exists, $value) = (1, $p->{schema}{default})
+      if !$exists
+      and $p->{schema}
+      and exists $p->{schema}{default};
+    ($exists, $value) = (1, $p->{default}) if !$exists and exists $p->{default};
+
+    $self->_coerce_input($type, $value) unless $in eq 'body';
+
     if (my @e = $self->_validate_request_value($p, $name => $value)) {
       push @errors, @e;
     }
@@ -153,9 +122,11 @@ sub validate_response {
   return JSON::Validator::E('/' => "No responses rules defined for status $status.")
     unless my $res_schema = $schema->{responses}{$status} || $schema->{responses}{default};
 
-  if ($self->version eq '3') {
+  local $self->{validate_response} = 1;
+
+  if ($self->version eq '3' and $res_schema->{content}) {
     my $accept = $self->_negotiate_accept_header($c, $res_schema);
-    return JSON::Validator::E('/' => "No responses rules defined for Accept $accept.")
+    return JSON::Validator::E('/' => "No responses rules defined for $accept.")
       unless $res_schema = $res_schema->{content}{$accept};
     $c->stash(openapi_negotiated_content_type => $accept);
   }
@@ -190,11 +161,32 @@ sub _build_formats {
   return $formats;
 }
 
+sub _coerce_input {
+  my ($self, $type) = @_;
+  return unless $type and defined $_[2];
+
+  if ($type ne 'array' and ref $_[2] eq 'ARRAY') {
+    $_[2] = $_[2]->[-1];
+  }
+  if (($type eq 'integer' or $type eq 'number') and Scalar::Util::looks_like_number($_[2])) {
+    $_[2] += 0;
+  }
+  elsif ($type eq 'boolean') {
+    if (!$_[2] or $_[2] =~ /^(?:false)$/) {
+      $_[2] = false;
+    }
+    elsif ($_[2] =~ /^(?:1|true)$/) {
+      $_[2] = true;
+    }
+  }
+}
+
 sub _coerce_by_collection_format {
   my ($self, $data, $p) = @_;
+  return $data unless defined $data;
 
-  my $schema = $p->{schema} || $p;
-  my $type = ($schema->{items} ? $schema->{items}{type} : $schema->{type}) || '';
+  my $schema = $p->{schema}                                                  || $p;
+  my $type   = ($schema->{items} ? $schema->{items}{type} : $schema->{type}) || '';
 
   my $collection_format = $p->{collectionFormat};
   my $custom_re;
@@ -205,11 +197,11 @@ sub _coerce_by_collection_format {
       $collection_format = 'csv';
     }
     elsif ($p->{style} eq 'label') {
-      $custom_re = qr{\.};
+      $custom_re         = qr{\.};
       $collection_format = $p->{explode} ? 'custom' : 'csv' if $data =~ s/^$custom_re//;
     }
     elsif ($p->{style} eq 'matrix') {
-      $custom_re = qr{;\Q$p->{name}\E=};
+      $custom_re         = qr{;\Q$p->{name}\E=};
       $collection_format = $p->{explode} ? 'custom' : 'csv' if $data =~ s/^$custom_re//;
     }
     elsif ($p->{style} eq 'form') {
@@ -226,14 +218,13 @@ sub _coerce_by_collection_format {
   return $data unless $collection_format;
 
   if ($collection_format eq 'multi') {
-    $data = [$data] unless ref $data eq 'ARRAY';
+    $data  = [$data] unless ref $data eq 'ARRAY';
     @$data = map { $_ + 0 } @$data if $type eq 'integer' or $type eq 'number';
     return $data;
   }
 
-  my $re = $collection_format eq 'custom' ? $custom_re
-                                          : $COLLECTION_RE{$collection_format} || ',';
-
+  my $re
+    = $collection_format eq 'custom' ? $custom_re : $self->_re_for_collection($collection_format);
   my $single = ref $data eq 'ARRAY' ? 0 : ($data = [$data]);
 
   for my $i (0 .. @$data - 1) {
@@ -244,10 +235,116 @@ sub _coerce_by_collection_format {
   return $single ? $data->[0] : $data;
 }
 
+sub _coerce_object_default_explode {
+  my ($self, $in) = @_;
+  return $in eq 'cookie' || $in eq 'query' ? true : false;
+}
+
+sub _coerce_object_default_style {
+  my ($self, $in) = @_;
+  return 'form'   if $in eq 'cookie' or $in eq 'query';
+  return 'simple' if $in eq 'header' or $in eq 'path';
+  return undef;
+}
+
+sub _coerce_object_by_style {
+  my ($self, $c, $data, $p) = @_;
+
+  my $style   = $p->{style}   // $self->_coerce_object_default_style($p->{in});
+  my $explode = $p->{explode} // $self->_coerce_object_default_explode($p->{in});
+  return $data unless $style;
+
+  # Special serializations
+  if ($style eq 'form' && $explode) {
+    return $c->req->url->query->to_hash;
+  }
+  elsif ($style eq 'deepObject') {
+    return $self->_coerce_deep_object($c, $p);
+  }
+
+  return unless defined $data;
+
+  if ($explode) {
+    return $data unless my $re = $self->_re_for_object_explode_true($style);
+    return if $style eq 'matrix' && $data !~ s/^;//;
+    return if $style eq 'label'  && $data !~ s/^\.//;
+    my $params = Mojo::Parameters->new;
+    $params->append(Mojo::Parameters->new($_)) for split /$re/, $data;
+    return $params->to_hash;
+  }
+  else {
+    return $data unless my $re = $self->_re_for_object_explode_false($style);
+    return if $style eq 'matrix' && $data !~ s/^;\Q$p->{name}\E=//;
+    return if $style eq 'label'  && $data !~ s/^\.//;
+    return Mojo::Parameters->new->pairs([split /$re/, $data])->to_hash;
+  }
+}
+
+sub _coerce_deep_object {
+  my ($self, $c, $p) = @_;
+
+  my (@pairs, $result) = @{$c->req->params->pairs};
+  while (my ($k, $v) = splice(@pairs, 0, 2)) {
+
+    # Retrieve the list of the deep keys
+    my @deep_keys;
+    push @deep_keys, $2 while ($k =~ s/^(\Q$p->{name}\E)\[([^]]*)\]/$1/);
+    next unless @deep_keys;
+
+    # Build the deep object
+    my $last_ref = \$result;
+    foreach my $key (@deep_keys) {
+      if ($key eq '' || $key =~ /^\d+$/) {
+        $$last_ref //= [];
+        $key = ($#{$$last_ref} + 1) if $key eq '';
+        $$last_ref->[$key] //= undef;
+        $last_ref = \$$last_ref->[$key];
+      }
+      else {
+        $$last_ref //= {};
+        $$last_ref->{$key} //= undef;
+        $last_ref = \$$last_ref->{$key};
+      }
+    }
+
+    # Set the value into the last reference
+    if (!defined $$last_ref) {
+      $$last_ref = $v;
+    }
+    elsif (ref $$last_ref eq 'ARRAY') {
+      push @{$$last_ref}, $v;
+    }
+    else {
+      $$last_ref = [$$last_ref, $v];
+    }
+  }
+
+  return $result;
+}
+
 sub _confess_invalid_in {
   confess "Unsupported \$in: $_[0]. Please report at https://github.com/jhthorsen/json-validator";
 }
 
+sub _definitions_path_for_ref {
+  my ($self, $ref) = @_;
+
+  if ($self->version eq '3') {
+
+    # Try to determine the path from the fqn
+    # We are only interested in the path in the fqn, so following fqn:
+    #
+    # #/components/schemas/some_schema, the returned path with be ['components', 'schemas']
+    my $path = Mojo::Path->new($ref->fqn =~ m!^.*#/(components/.+)$!)->to_dir->parts;
+    return $path->[0] ? $path : ['definitions'];
+  }
+  else {
+
+    # By default return definitions as path
+    return ['definitions'];
+  }
+}
+
 sub _get_request_data {
   my ($self, $c, $in) = @_;
 
@@ -258,7 +355,7 @@ sub _get_request_data {
     return $c->match->stack->[-1];
   }
   elsif ($in eq 'formData') {
-    return $self->{cache}{$in} ||= $c->req->body_params->to_hash(1);
+    return $self->{cache}{$in} ||= $self->_get_request_form_and_uploads($c);
   }
   elsif ($in eq 'cookie') {
     return $self->{cache}{$in} ||= {map { ($_->name, $_->value) } @{$c->req->cookies}};
@@ -275,6 +372,14 @@ sub _get_request_data {
   }
 }
 
+sub _get_request_form_and_uploads {
+  my ($self, $c) = @_;
+  my $form = $c->req->body_params->to_hash(1);
+  push @{$form->{$_->name}}, $_->size ? 'x' : ''
+    for grep { !exists $form->{$_->name} } @{$c->req->uploads};
+  return $form;
+}
+
 sub _get_request_uploads {
   my ($self, $c, $name) = @_;
   return $c->req->every_upload($name);
@@ -292,7 +397,7 @@ sub _match_number {
   my ($name, $val, $format) = @_;
   return 'Does not look like an integer' if $name =~ m!^int! and $val !~ /^-?\d+(\.\d+)?$/;
   return 'Does not look like a number.' unless looks_like_number $val;
-  return undef unless $format;
+  return undef                          unless $format;
   return undef if $val eq unpack $format, pack $format, $val;
   return "Does not match $name format.";
 }
@@ -328,10 +433,36 @@ sub _negotiate_accept_header {
   return $accept;
 }
 
+sub _re_for_collection {
+  my ($self, $style) = @_;
+  return qr{,}  if $style eq 'csv';
+  return qr{\s} if $style eq 'ssv';
+  return qr{\t} if $style eq 'tsv';
+  return qr{\|} if $style eq 'pipes';
+  return ',';
+}
+
+sub _re_for_object_explode_false {
+  my ($self, $style) = @_;
+  return qr{,}  if $style eq 'form' or $style eq 'matrix' or $style eq 'simple';
+  return qr{\.} if $style eq 'label';
+  return qr{\|} if $style eq 'pipeDelimited';
+  return qr{\s} if $style eq 'spaceDelimited';
+  return undef;
+}
+
+sub _re_for_object_explode_true {
+  my ($self, $style) = @_;
+  return qr{\.} if $style eq 'label';
+  return qr{;}  if $style eq 'matrix';
+  return qr{,}  if $style eq 'simple';
+  return undef;
+}
+
 sub _resolve_ref {
-  my ($self, $topic, $url) = @_;
-  $topic->{'$ref'} = "#/definitions/$topic->{'$ref'}" if $topic->{'$ref'} =~ /^\w+$/;
-  return $self->SUPER::_resolve_ref($topic, $url);
+  my ($self, $ref, $base, $root) = @_;
+  $ref = "#/definitions/$ref" if $ref =~ /^\w+$/;
+  return $self->SUPER::_resolve_ref($ref, $base, $root);
 }
 
 sub _set_request_data {
@@ -368,14 +499,19 @@ sub _validate_request_body {
   my $ct = $c->req->headers->content_type // '';
 
   $ct =~ s!;.*$!!;
-  if (my $schema = $body_schema->{content}{$ct}) {
+  if (my $content = $body_schema->{content}{$ct}) {
     my $body = $self->_get_request_data($c, $ct =~ /\bform\b/ ? 'formData' : 'body');
-    local $schema->{required} //= $body_schema->{required};
-    return $self->_validate_request_value($schema, body => $body);
+    local $content->{required} //= $body_schema->{required};
+    if (ref $content->{schema} eq 'HASH' and ref $body eq 'HASH') {
+      $self->_coerce_input($content->{schema}{properties}{$_}{type}, $body->{$_})
+        for keys %{$content->{schema}{properties} || {}};
+    }
+    return $self->_validate_request_value($content, body => $body);
   }
 
   return JSON::Validator::E('/' => "No requestBody rules defined for Content-Type $ct.") if $ct;
-  return JSON::Validator::E('/', 'Invalid Content-Type.');
+  return JSON::Validator::E('/', 'Invalid Content-Type.') if $body_schema->{required};
+  return;
 }
 
 sub _validate_request_value {
@@ -466,18 +602,26 @@ sub _validate_type_object {
   if ($self->version eq '3') {
     for my $key (keys %properties) {
       next unless $properties{$key}{nullable};
-      $properties{$key} = {%{$properties{$key}}};
+      my $tied = tied %{$properties{$key}};
+      $properties{$key} = $tied ? {%{$tied->schema}} : {%{$properties{$key}}};
       $properties{$key}{type} = ['null', _to_list($properties{$key}{type})];
     }
   }
 
-  return shift->SUPER::_validate_type_object(@_) unless $self->{validate_input};
+  return shift->_validate_type_object_input(@_)    if $self->{validate_input};
+  return shift->_validate_type_object_response(@_) if $self->{validate_response};
+  return shift->SUPER::_validate_type_object(@_);
+}
+
+sub _validate_type_object_input {
+  my ($self, $data, $path, $schema) = @_;
+  my $properties = $schema->{properties};  # OK to modify, since called from _validate_type_object()
 
   my (@e, %ro);
-  for my $p (keys %properties) {
-    next unless $properties{$p}{readOnly};
-    push @e, JSON::Validator::E("$path/$p", "Read-only.") if exists $data->{$p};
-    $ro{$p} = 1;
+  for my $name (keys %$properties) {
+    next unless $properties->{$name}{readOnly};
+    push @e, JSON::Validator::E("$path/$name", "Read-only.") if exists $data->{$name};
+    $ro{$name} = 1;
   }
 
   my $discriminator = $schema->{discriminator};
@@ -495,6 +639,23 @@ sub _validate_type_object {
   return @e, $self->SUPER::_validate_type_object($data, $path, $schema);
 }
 
+sub _validate_type_object_response {
+  my ($self, $data, $path, $schema) = @_;
+  my $properties = $schema->{properties};  # OK to modify, since called from _validate_type_object()
+
+  my %wo;
+  for my $name (keys %$properties) {
+    next unless $properties->{$name}{writeOnly};
+    delete $properties->{$name};
+    $wo{$name} = 1;
+  }
+
+  local $schema->{required} = $schema->{required} || [];
+  $schema->{required} = [grep !$wo{$_}, @{$schema->{required}}];
+
+  return $self->SUPER::_validate_type_object($data, $path, $schema);
+}
+
 1;
 
 =encoding utf8
@@ -651,10 +812,6 @@ OpenAPI schema.
 
 =back
 
-=head2 generate_definitions_path
-
-See L<JSON::Validator/generate_definitions_path>.
-
 =head2 validate_input
 
   @errors = $validator->validate_input($data, $schema);
