@@ -7,7 +7,7 @@ use Mojo::JSON;
 use Mojo::Util;
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '2.21';
+our $VERSION = '3.41';
 my $X_RE = qr{^x-};
 
 has route     => sub {undef};
@@ -38,7 +38,7 @@ sub register {
 
   unless ($app->defaults->{'openapi.base_paths'}) {
     $app->helper('openapi.spec'        => \&_helper_get_spec);
-    $app->helper('openapi.valid_input' => sub { _helper_validate($_[0]) ? undef : $_[0] });
+    $app->helper('openapi.valid_input' => \&_helper_valid_input);
     $app->helper('openapi.validate'    => \&_helper_validate);
     $app->helper('reply.openapi'       => \&_helper_reply);
     $app->hook(before_render => \&_before_render);
@@ -50,7 +50,7 @@ sub register {
 
   $self->{default_response_codes} = $config->{default_response_codes} || [400, 401, 404, 500, 501];
   $self->{default_response_name}  = $config->{default_response_name}  || 'DefaultResponse';
-  $self->{log_level} = $ENV{MOJO_OPENAPI_LOG_LEVEL} || $config->{log_level} || 'warn';
+  $self->{log_level}              = $ENV{MOJO_OPENAPI_LOG_LEVEL} || $config->{log_level} || 'warn';
   $self->_renderer($config->{renderer}) if $config->{renderer};
   $self->_build_route($app, $config);
 
@@ -58,7 +58,7 @@ sub register {
   for my $plugin (@{$config->{plugins} || [qw(+Cors +SpecRenderer +Security)]}) {
     $plugin = "Mojolicious::Plugin::OpenAPI::$plugin" if $plugin =~ s!^\+!!;
     eval "require $plugin;1" or Carp::confess("require $plugin: $@");
-    push @plugins, $plugin->new->register($app, $self, $config);
+    push @plugins, $plugin->new->register($app, {%$config, openapi => $self});
   }
 
   $self->_add_routes($app, $config);
@@ -98,51 +98,43 @@ sub _add_routes {
   my ($self, $app, $config) = @_;
   my (@routes, %uniq);
 
-  my @sorted_openapi_paths
-    = map { $_->[0] }
-    sort { $a->[1] <=> $b->[1] || length $a->[0] <=> length $b->[0] }
-    map { [$_, /\{/ ? 1 : 0] } grep { !/$X_RE/ } keys %{$self->validator->get('/paths') || {}};
-
-  for my $openapi_path (@sorted_openapi_paths) {
+  $self->_each_route(sub {
+    my ($http_method, $openapi_path) = @_;
     my $path_parameters = $self->validator->get([paths => $openapi_path => 'parameters']) || [];
+    my $op_spec         = $self->validator->get([paths => $openapi_path => $http_method]);
+    my $name            = $op_spec->{'x-mojo-name'} || $op_spec->{operationId};
+    my $to              = $op_spec->{'x-mojo-to'};
+    my $r;
 
-    for my $http_method (sort keys %{$self->validator->get([paths => $openapi_path]) || {}}) {
-      next if $http_method =~ $X_RE or $http_method eq 'parameters';
-      my $op_spec = $self->validator->get([paths => $openapi_path => $http_method]);
-      my $name    = $op_spec->{'x-mojo-name'} || $op_spec->{operationId};
-      my $to      = $op_spec->{'x-mojo-to'};
-      my $r;
+    $self->{parameters_for}{$openapi_path}{$http_method}
+      = [@$path_parameters, @{$op_spec->{parameters} || []}];
 
-      $self->{parameters_for}{$openapi_path}{$http_method}
-        = [@$path_parameters, @{$op_spec->{parameters} || []}];
+    die qq([OpenAPI] operationId "$op_spec->{operationId}" is not unique)
+      if $op_spec->{operationId} and $uniq{o}{$op_spec->{operationId}}++;
+    die qq([OpenAPI] Route name "$name" is not unique.) if $name and $uniq{r}{$name}++;
 
-      die qq([OpenAPI] operationId "$op_spec->{operationId}" is not unique)
-        if $op_spec->{operationId} and $uniq{o}{$op_spec->{operationId}}++;
-      die qq([OpenAPI] Route name "$name" is not unique.) if $name and $uniq{r}{$name}++;
-
-      if (!$to and $name) {
-        $r = $self->route->root->find($name);
-        warn "[OpenAPI] Found existing route by name '$name'.\n" if DEBUG and $r;
-        $self->route->add_child($r) if $r;
-      }
-      if (!$r) {
-        my $route_path = $self->_openapi_path_to_route_path($http_method, $openapi_path);
-        $name ||= $op_spec->{operationId};
-        warn "[OpenAPI] Creating new route for '$route_path'.\n" if DEBUG;
-        $r = $self->route->$http_method($route_path);
-        $r->name("$self->{route_prefix}$name") if $name;
-      }
-
-      $self->_add_default_response($op_spec);
-
-      $r->to(ref $to eq 'ARRAY' ? @$to : $to) if $to;
-      $r->to({'openapi.method' => $http_method});
-      $r->to({'openapi.path'   => $openapi_path});
-      warn "[OpenAPI] Add route $http_method @{[$r->to_string]} (@{[$r->name // '']})\n" if DEBUG;
-
-      push @routes, $r;
+    if (!$to and $name) {
+      $r = $self->route->root->find($name);
+      warn "[OpenAPI] Found existing route by name '$name'.\n" if DEBUG and $r;
+      $self->route->add_child($r) if $r;
     }
-  }
+    if (!$r) {
+      my $route_path = $self->_openapi_path_to_route_path($http_method, $openapi_path);
+      $name ||= $op_spec->{operationId};
+      warn "[OpenAPI] Creating new route for '$route_path'.\n" if DEBUG;
+      $r = $self->route->$http_method($route_path);
+      $r->name("$self->{route_prefix}$name") if $name;
+    }
+
+    $self->_add_default_response($op_spec);
+
+    $r->to(ref $to eq 'ARRAY' ? @$to : $to) if $to;
+    $r->to({'openapi.method' => $http_method});
+    $r->to({'openapi.path'   => $openapi_path});
+    warn "[OpenAPI] Add route $http_method @{[$r->to_string]} (@{[$r->name // '']})\n" if DEBUG;
+
+    push @routes, $r;
+  });
 
   $app->plugins->emit_hook(openapi_routes_added => $self, \@routes);
 }
@@ -228,6 +220,22 @@ sub _default_schema_v3 {
   };
 }
 
+sub _each_route {
+  my ($self, $cb) = @_;
+
+  my @sorted_paths
+    = map { $_->[0] }
+    sort  { $a->[1] <=> $b->[1] || length $a->[0] <=> length $b->[0] }
+    map   { [$_, /\{/ ? 1 : 0] } grep { !/$X_RE/ } keys %{$self->validator->get('/paths') || {}};
+
+  my @routes;
+  for my $path (@sorted_paths) {
+    for my $http_method (sort keys %{$self->validator->get([paths => $path]) || {}}) {
+      $cb->($http_method => $path) unless $http_method =~ $X_RE or $http_method eq 'parameters';
+    }
+  }
+}
+
 sub _helper_get_spec {
   my $c    = shift;
   my $path = shift // 'for_current';
@@ -268,27 +276,22 @@ sub _helper_reply {
   return $c->render(@args, openapi => $output);
 }
 
+sub _helper_valid_input {
+  my $c = shift;
+  return undef if $c->res->code;
+  return $c unless my @errors = _helper_validate($c);
+  $c->stash(status => 400)
+    ->render(data => _self($c)->_renderer->($c, {errors => \@errors, status => 400}));
+  return undef;
+}
+
 sub _helper_validate {
-  my ($c, $args) = @_;
-
-  # code() can be set by other methods such as $c->openapi->cors_simple()
-  return [{message => 'Already rendered.'}] if $c->res->code;
-
-  # Write validated data to $c->validation->output
+  my $c       = shift;
   my $self    = _self($c);
   my $op_spec = $c->openapi->spec;
   local $op_spec->{parameters}
     = $self->_parameters_for($c->req->method, $c->stash('openapi.path'),);
-  my @errors = $self->validator->validate_request($c, $op_spec, $c->validation->output);
-
-  if (@errors) {
-    $self->_log($c, '<<<', \@errors);
-    $c->stash(status => 400)
-      ->render(data => $self->_renderer->($c, {errors => \@errors, status => 400}))
-      if $args->{auto_render} // 1;
-  }
-
-  return @errors;
+  return $self->validator->validate_request($c, $op_spec, $c->validation->output);
 }
 
 sub _log {
@@ -372,7 +375,7 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
     my $c = shift->openapi->valid_input or return;
 
     # Generate some data
-    my $data = {body => $c->validation->param("body")};
+    my $data = {body => $c->req->json};
 
     # Validate the output response and render it to the user agent
     # using a custom "openapi" handler.
@@ -380,7 +383,9 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
   }, "echo";
 
   # Load specification and start web server
-  plugin OpenAPI => {url => "data:///spec.json"};
+  # Use "v3" instead of "v2" for "schema" if you are using OpenAPI v3
+  # The plugin must be loaded *after* defining the routes in a Lite app
+  plugin OpenAPI => {url => "data:///spec.json", schema => "v2"};
   app->start;
 
   __DATA__
@@ -408,20 +413,28 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
     }
   }
 
-See L<Mojolicious::Plugin::OpenAPI::Guides::Tutorial> for a tutorial on how to
+See L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv2> or
+L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv3> for tutorials on how to
 write a "full" app with application class and controllers.
 
+=head1 IMPORTANT ANNOUNCEMENT
+
+Next version of L<Mojolicious::Plugin::OpenAPI> will not be shipped with
+L<JSON::Validator::OpenAPI::Mojolicious>. Instead, it will depend on the new
+L<JSON::Validator::Schema::OpenAPIv2> and L<JSON::Validator::Schema::OpenAPIv3>
+modules.
+
+The next version is available at
+L<https://github.com/jhthorsen/mojolicious-plugin-openapi/pull/160>.
+
 =head1 DESCRIPTION
 
 L<Mojolicious::Plugin::OpenAPI> is L<Mojolicious::Plugin> that add routes and
 input/output validation to your L<Mojolicious> application based on a OpenAPI
-(Swagger) specification.
+(Swagger) specification. This plugin supports both version L<2.0|/schema> and
+L<3.x|/schema>, though 3.x I<might> have some missing features.
 
-Have a look at the L</SEE ALSO> for references to more documentation, or jump
-right to the L<tutorial|Mojolicious::Plugin::OpenAPI::Guides::Tutorial>.
-
-Currently v2 is very well supported, while v3 should be considered
-EXPERIMENTAL.
+Have a look at the L</SEE ALSO> for references to more documentation.
 
 Please report in L<issues|https://github.com/jhthorsen/json-validator/issues>
 or open pull requests to enhance the 3.0 support.
@@ -458,16 +471,6 @@ L<JSON::Validator::Error> objects or empty list on valid input.
 Note that this helper is only for customization. You probably want
 L</openapi.valid_input> in most cases.
 
-Validated input parameters will be copied to
-C<Mojolicious::Controller/validation>, which again can be extracted by the
-"name" in the parameters list from the spec. Example:
-
-  # specification:
-  "parameters": [{"in": "body", "name": "whatever", "schema": {"type": "object"}}],
-
-  # controller
-  my $body = $c->validation->param("whatever");
-
 =head2 openapi.valid_input
 
   $c = $c->openapi->valid_input;
@@ -581,7 +584,7 @@ plugin.
 
 The name of the "definition" in the spec that will be used for
 L</default_response_codes>. The default value is "DefaultResponse". See
-L<Mojolicious::Plugin::OpenAPI::Guides::Tutorial/"Default response schema">
+L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv2/"Default response schema">
 for more details.
 
 =head3 log_level
@@ -621,6 +624,9 @@ C<route> can be specified in case you want to have a protected API. Example:
 Can be used to set a different schema, than the default OpenAPI 2.0 spec.
 Example values: "http://swagger.io/v2/schema.json", "v2" or "v3".
 
+See also L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv2> and
+L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv3>.
+
 =head3 spec_route_name
 
 Name of the route that handles the "basePath" part of the specification and
@@ -665,15 +671,29 @@ the terms of the Artistic License version 2.0.
 
 =over 2
 
-=item * L<Mojolicious::Plugin::OpenAPI::Guides::Tutorial>
+=item * L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv2>
+
+Guide for how to use this plugin with OpenAPI version 2.0 spec.
+
+=item * L<Mojolicious::Plugin::OpenAPI::Guides::OpenAPIv3>
+
+Guide for how to use this plugin with OpenAPI version 3.0 spec.
 
 =item * L<Mojolicious::Plugin::OpenAPI::Cors>
 
+Plugin to add Cross-Origin Resource Sharing (CORS).
+
 =item * L<Mojolicious::Plugin::OpenAPI::Security>
 
+Plugin for handling security definitions in your schema.
+
 =item * L<Mojolicious::Plugin::OpenAPI::SpecRenderer>
 
-=item * L<OpenAPI specification|https://openapis.org/specification>
+Plugin for exposing your spec in human readble or JSON format.
+
+=item * L<https://www.openapis.org/>
+
+Official OpenAPI website.
 
 =back
 
